"""Details of the Gwyddion Simple Field format."""

import numpy as np

null_byte = b"\x00"  # bytes
null_char = null_byte.decode()  # str
gsf_magic_line = "Gwyddion Simple Field 1.0\n"
gsf_known_metadata_types = {
    "XRes": int,
    "YRes": int,
    "XReal": float,
    "YReal": float,
    "XOffset": float,
    "YOffset": float,
    "Title": str,
    "XYUnits": str,
    "ZUnits": str,
}
gsf_known_metadata_order = {
    key: i for i, key in enumerate(gsf_known_metadata_types, start=1)
}
# 32-bit (4-bytes) little-endian floats.
gsf_dtype = np.dtype("<f4")
# Row-major order.
gsf_array_order = "C"


def gsf_padding_lenght(header_length):
    """Length of the null-byte padding between metadata and data."""
    return 4 - header_length % 4



"""Read Gwyddion Simple Field files."""

from contextlib import suppress
from pathlib import Path

import numpy as np

from .format import (
    gsf_dtype,
    gsf_known_metadata_types,
    gsf_magic_line,
    gsf_padding_lenght,
    null_byte,
)


def read_gsf(
    path: Path | str,
) -> tuple[np.typing.NDArray[np.float32], dict[str, float | str]]:
    """Read a Gwyddion Simple Field file (.gsf).

    Parameters
    ----------
        path
            Path to the file to be read.

    Returns
    -------
        metadata
            A dict of metadata. The fields XRes and YRes are not included,
            since they would be a duplicate of data.shape. Custom fields not mentioned
            in the Gwyddion Simple Field specification are read as strings.
        data
            A 2-dimensional NumPy array of float32.

    Raises
    ------
        ValueError
            If the file to be read is not a Gwyddion Simple Field.
        KeyError
            If required metadata is missing from the file to be read.
    """
    path = Path(path)
    metadata = {}

    with path.open("rb") as file:
        # Check magic line.
        if file.readline().decode() != gsf_magic_line:
            msg = f"Magic line not found at the beginning of {path}"
            raise ValueError(msg)

        # Read metadata.
        # Peek does not do what you think it does.
        # https://stackoverflow.com/a/24474743
        while file.peek(1)[:1] != null_byte:
            key, value = file.readline().decode().split("=")
            metadata[key.strip()] = value.strip()
        for key, type_ in gsf_known_metadata_types.items():
            with suppress(KeyError):
                metadata[key] = type_(metadata[key])

        # Skip null-byte padding.
        header_length = file.tell()
        file.seek(gsf_padding_lenght(header_length), 1)

        # Read data.
        shape = metadata["YRes"], metadata["XRes"]
        data_size = gsf_dtype.itemsize * shape[0] * shape[1]
        data = np.frombuffer(file.read(data_size), dtype=gsf_dtype).reshape(shape)

        if file.read(1) != b"":
            msg = f"Unexpected additional data found at the end of {path}"
            raise ValueError(msg)

    # Do not duplicate information already present in data.shape.
    del metadata["XRes"]
    del metadata["YRes"]

    return data, metadata



"""Write Gwyddion Simple Field files."""

from pathlib import Path

import numpy as np

from .format import (
    gsf_array_order,
    gsf_dtype,
    gsf_known_metadata_order,
    gsf_magic_line,
    gsf_padding_lenght,
    null_byte,
    null_char,
)


def write_gsf(
    path: Path | str,
    data: np.typing.NDArray[np.float32],
    metadata: dict | None = None,
):
    """Write a NumPy array to a Gwyddion Simple Field file (.gsf).

    Parameters
    ----------
        path
            Path to the output file to be written, with .gsf extension.
        data
            A 2-dimensional array of float32. 0- and 1-dimensional arrays
            will be reshaped to 2 dimensions. Inf and NaN are supported.
        metadata : optional
            Additional metadata to be included in the output file. The optional fields
            defined by the Gwyddion Simple Field format will be written first, followed
            by the custom fields defined by the user.

    Raises
    ------
        ValueError
            If the input parameters are not compatible with the Gwyddion Simple Field
            file format.
    """
    data = prepare_data(data)
    if data.ndim >= 3:  # noqa: PLR2004
        msg = (
            f"data.shape is {data.shape}, but the Gwyddion Simple Field file format "
            "supports at most 2-dimensional data."
        )
        raise ValueError(msg)
    if data.dtype != np.float32:
        msg = (
            f"data.dtype is {data.dtype}, but the Gwyddion Simple Field file format "
            "supports only 32-bit floating point data. "
            "Convert with data.astype(np.float32)."
        )
        raise ValueError(msg)

    path = Path(path)
    if path.suffix != ".gsf":
        msg = "The Gwyddion Simple Field file format uses the .gsf file name extension."
        raise ValueError(msg)

    metadata = prepare_metadata(metadata)
    if ("XRes" in metadata) or ("YRes" in metadata):
        msg = (
            "Do not specify neither XRes nor YRes in metadata. "
            "They are derived form data.shape."
        )
        raise ValueError(msg)
    for key, value in metadata.items():
        if ("=" in key) or ("=" in value):
            msg = (
                "Equal sign '=' not allowed in metadata. "
                f"Offending key: {key}, offending value: {value}"
            )
            raise ValueError(msg)
        if (null_char in key) or (null_char in value):
            msg = (
                "Null character (ASCII code 0) not allowed in metadata. "
                f"Offending key: {key}, offending value: {value}"
            )
            raise ValueError(msg)
    metadata = {"XRes": data.shape[1], "YRes": data.shape[0]} | metadata

    header = gsf_magic_line
    for key, value in metadata.items():
        header += f"{key} = {value}\n"
    header_bytes = bytes(header, "utf-8")
    gsf_padding = null_byte * gsf_padding_lenght(len(header_bytes))

    with path.open("wb") as file:
        file.write(header_bytes)
        file.write(gsf_padding)
        file.write(data.astype(gsf_dtype).tobytes(order=gsf_array_order))


def prepare_data(data):
    """Reshape 0- and 1-dimensional arrays to 2 dimensions."""
    return np.atleast_2d(data)


def prepare_metadata(metadata):
    """Sort, convert to string, and strip whitespace from the metadata."""
    if metadata is None:
        metadata = {}
    # Put the known metadata first, in their canonical order,
    # then the custom metadata, in insertion order.
    metadata = dict(
        sorted(
            metadata.items(),
            key=lambda item: gsf_known_metadata_order.get(item[0], np.inf),
        )
    )
    metadata = {str(key).strip(): str(value).strip() for key, value in metadata.items()}
    return metadata  # noqa: RET504



"""Read and write Gwyddion Simple Field files (.gsf).

The `Gwyddion Simple Field <http://gwyddion.net/documentation/user-guide-en/gsf.html>`_
file format (.gsf) is a single-channel format for 2D data that was designed to be
easy and efficient to read and write, with human-readable header, reasonably expressive,
and avoiding instrument or application specific fields
(though it can optionally bear them).
"""

from importlib.metadata import version

from .read import read_gsf
from .write import write_gsf

__version__ = version("gsffile")
__all__ = ["read_gsf", "write_gsf"]



"""Test that what is read is what we expect, given what was written."""

from pathlib import Path

import numpy as np
import pytest

from gsffile import read_gsf, write_gsf
from gsffile.write import prepare_data, prepare_metadata

rng = np.random.default_rng(seed=1)


def assert_roundtrip_ok(tmp_path, data, meta=None):
    """Assert that what is read is what we expect, given what was written."""
    path = tmp_path / "test.gsf"
    write_gsf(path, data, meta)
    data_2, meta_2 = read_gsf(path)
    expected_data = prepare_data(data)
    expected_meta = prepare_metadata(meta)
    assert meta_2 == expected_meta
    # Check metadata order.
    assert list[meta_2.keys()] == list[expected_meta.keys()]
    np.testing.assert_array_equal(data_2, expected_data)


allowed_shapes = [(3, 2), (3, 1), (1, 3), (3,), (1,), (0,), (), (0, 0)]


@pytest.mark.parametrize("shape", allowed_shapes, ids=lambda shape: f"shape {shape}")
def test_allowed_shapes(tmp_path: Path, shape: tuple[int]):
    """Test 0-, 1-, and 2-dimensional random data."""
    data = rng.uniform(size=shape).astype(np.float32)
    assert_roundtrip_ok(tmp_path, data)


def test_meta(tmp_path: Path):
    """Test metadata roundtrip."""
    # Optional fields of non-str type, such as XReal, would fail this test.
    data = np.zeros((2, 3), dtype=np.float32)
    meta = {
        "string": "a string",
        "float": np.sqrt(3),
        "int": -0,
        "  whitespace  ": " ",
        "multi-byte characters": "💣",
        "ZUnits": "m",
        "XYUnits": "m",
    }
    assert_roundtrip_ok(tmp_path, data, meta)


def test_non_finite(tmp_path: Path):
    """Test roundtrip for inf and NaN."""
    data = np.array([[np.inf, -np.inf], [np.nan, -0.0]], dtype=np.float32)
    assert_roundtrip_ok(tmp_path, data)


def test_path_as_str(tmp_path: Path):
    """Test passing a str as path."""
    data = np.zeros((2, 3), dtype=np.float32)
    path = tmp_path / "test.gsf"
    write_gsf(str(path), data)
    _ = read_gsf(str(path))



"""Test read_gsf."""

from pathlib import Path

import pytest

from gsffile import read_gsf


def test_missing_magic_line(tmp_path: Path):
    """Test missing XRes metadata."""
    content = "Gwyddion Simple Field\nXRes = 1\nYRes = 1\n\x00\x00\x00\x00_¨Ÿ>"
    path = tmp_path / "test.gsf"
    path.write_bytes(bytes(content, encoding="utf-8"))
    with pytest.raises(ValueError, match=r"Magic line not found.*"):
        _ = read_gsf(path)


def test_missing_XRes(tmp_path: Path):  # noqa: N802
    """Test missing XRes metadata."""
    content = "Gwyddion Simple Field 1.0\nYRes = 1\n\x00\x00\x00\x00_¨Ÿ>"
    path = tmp_path / "test.gsf"
    path.write_bytes(bytes(content, encoding="utf-8"))
    with pytest.raises(KeyError):
        _ = read_gsf(path)


def test_additional_data(tmp_path: Path):
    """Test additional data at the end of the file."""
    content = (
        "Gwyddion Simple Field 1.0\n"
        "XRes = 1\n"
        "YRes = 1\n"
        "\x00\x00\x00\x00_¨Ÿ>"
        "additional_data"
    )
    path = tmp_path / "test.gsf"
    path.write_bytes(bytes(content, encoding="utf-8"))
    with pytest.raises(ValueError, match=r"Unexpected additional data.*"):
        _ = read_gsf(path)



"""Test write_gsf."""

from pathlib import Path

import numpy as np
import pytest

from gsffile.write import write_gsf


def test_wrong_extension(tmp_path: Path):
    """Test wrong file name extension --> ☠️."""
    data = np.zeros((3, 2), dtype=np.float32)
    with pytest.raises(ValueError, match=r".* .gsf file name extension.*"):
        write_gsf(tmp_path / "test.wrong_extension", data)


not_allowed_shapes = [(2, 3, 4), (2, 2, 1), (1, 2, 2), (1, 1, 1, 1), (0, 0, 0)]


@pytest.mark.parametrize(
    "shape", not_allowed_shapes, ids=lambda shape: f"shape {shape}"
)
def test_wrong_ndim(tmp_path: Path, shape: tuple[int]):
    """Test wrong ndim --> ☠️."""
    data = np.zeros(shape, dtype=np.float32)
    with pytest.raises(ValueError, match=r".* at most 2-dimensional data.*"):
        write_gsf(tmp_path / "test.gsf", data)


def test_wrong_dtype(tmp_path: Path):
    """Test wrong dtype --> ☠️."""
    data = np.zeros((3, 2), dtype=np.float64)
    with pytest.raises(ValueError, match=r".* only 32-bit floating point data.*"):
        write_gsf(tmp_path / "test.gsf", data)


@pytest.mark.parametrize(
    "meta", [{"XRes": 101}, {" XRes ": 101}], ids=lambda meta: next(iter(meta.keys()))
)
def test_XRes_in_meta(tmp_path: Path, meta: dict):  # noqa: N802
    """Test XRes in meta --> ☠️."""
    data = np.zeros((3, 2), dtype=np.float32)
    with pytest.raises(ValueError, match=r".* neither XRes nor YRes in metadata.*"):
        write_gsf(tmp_path / "test.gsf", data, meta)


@pytest.mark.parametrize(
    "meta", [{"key=key": "value"}, {"key": "="}], ids=lambda meta: str(meta)
)
def test_equal_sign_in_meta(tmp_path: Path, meta: dict):
    """Test equal sign in meta --> ☠️."""
    data = np.zeros((3, 2), dtype=np.float32)
    with pytest.raises(ValueError, match=r"Equal sign .*"):
        write_gsf(tmp_path / "test.gsf", data, meta)


@pytest.mark.parametrize(
    "meta",
    [{"key\x00key": "value"}, {"key": "value\x00value"}],
    ids=lambda meta: str(meta),
)
def test_null_char_in_meta(tmp_path: Path, meta: dict):
    """Test equal sign in meta --> ☠️."""
    data = np.zeros((3, 2), dtype=np.float32)
    with pytest.raises(ValueError, match=r"Null character .*"):
        write_gsf(tmp_path / "test.gsf", data, meta)
